@@ -1,13 +1,13 @@
 /*
    Find file command for the Midnight Commander
 
-   Copyright (C) 1995-2020
+   Copyright (C) 1995-2024
    Free Software Foundation, Inc.
 
    Written  by:
    Miguel de Icaza, 1995
    Slava Zanko <slavazanko@gmail.com>, 2013
-   Andrew Borodin <aborodin@vmail.ru>, 2013
+   Andrew Borodin <aborodin@vmail.ru>, 2013-2022
 
    This file is part of the Midnight Commander.
 
@@ -36,7 +36,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
-#include <sys/time.h>
 
 #include "lib/global.h"
 
@@ -55,7 +54,6 @@
 
 #include "dir.h"
 #include "cmd.h"                /* find_cmd(), view_file_at_line() */
-#include "midnight.h"           /* current_panel */
 #include "boxes.h"
 #include "panelize.h"
 
@@ -92,6 +90,7 @@ typedef struct
     gboolean file_case_sens;
     gboolean file_pattern;
     gboolean find_recurs;
+    gboolean follow_symlinks;
     gboolean skip_hidden;
     gboolean file_all_charsets;
 
@@ -115,13 +114,15 @@ typedef struct
     gsize end;
 } find_match_location_t;
 
-/*** file scope variables ************************************************************************/
+/*** forward declarations (file scope functions) *************************************************/
 
 /* button callbacks */
 static int start_stop (WButton * button, int action);
 static int find_do_view_file (WButton * button, int action);
 static int find_do_edit_file (WButton * button, int action);
 
+/*** file scope variables ************************************************************************/
+
 /* Parsed ignore dirs */
 static char **find_ignore_dirs = NULL;
 
@@ -134,6 +135,7 @@ static WLabel *content_label;   /* 'Content:' label */
 static WCheck *file_case_sens_cbox;     /* "case sensitive" checkbox */
 static WCheck *file_pattern_cbox;       /* File name is glob or regexp */
 static WCheck *recursively_cbox;
+static WCheck *follow_sym_cbox;
 static WCheck *skip_hidden_cbox;
 static WCheck *content_case_sens_cbox;  /* "case sensitive" checkbox */
 static WCheck *content_regexp_cbox;     /* "find regular expression" checkbox */
@@ -155,7 +157,7 @@ static unsigned long matches;   /* Number of matches */
 static gboolean is_start = FALSE;       /* Status of the start/stop toggle button */
 static char *old_dir = NULL;
 
-static struct timeval last_refresh;
+static gint64 last_refresh;
 
 /* Where did we stop */
 static gboolean resuming;
@@ -203,7 +205,7 @@ static const size_t quit_button = 4;    /* index of "Quit" button */
 static WListbox *find_list;     /* Listbox with the file list */
 
 static find_file_options_t options = {
-    TRUE, TRUE, TRUE, FALSE, FALSE,
+    TRUE, TRUE, TRUE, FALSE, FALSE, FALSE,
     TRUE, FALSE, FALSE, FALSE, FALSE,
     FALSE, NULL
 };
@@ -217,7 +219,7 @@ static mc_search_t *search_content_handle = NULL;
 /*** file scope functions ************************************************************************/
 /* --------------------------------------------------------------------------------------------- */
 
-/* don't use max macro to avoid double str_term_width1() call in widget length caclulation */
+/* don't use max macro to avoid double str_term_width1() call in widget length calculation */
 #undef max
 
 static int
@@ -288,6 +290,8 @@ find_load_options (void)
         mc_config_get_bool (mc_global.main_config, "FindFile", "file_shell_pattern", TRUE);
     options.find_recurs =
         mc_config_get_bool (mc_global.main_config, "FindFile", "file_find_recurs", TRUE);
+    options.follow_symlinks =
+        mc_config_get_bool (mc_global.main_config, "FindFile", "follow_symlinks", FALSE);
     options.skip_hidden =
         mc_config_get_bool (mc_global.main_config, "FindFile", "file_skip_hidden", FALSE);
     options.file_all_charsets =
@@ -321,6 +325,8 @@ find_save_options (void)
     mc_config_set_bool (mc_global.main_config, "FindFile", "file_shell_pattern",
                         options.file_pattern);
     mc_config_set_bool (mc_global.main_config, "FindFile", "file_find_recurs", options.find_recurs);
+    mc_config_set_bool (mc_global.main_config, "FindFile", "follow_symlinks",
+                        options.follow_symlinks);
     mc_config_set_bool (mc_global.main_config, "FindFile", "file_skip_hidden", options.skip_hidden);
     mc_config_set_bool (mc_global.main_config, "FindFile", "file_all_charsets",
                         options.file_all_charsets);
@@ -349,6 +355,14 @@ add_to_list (const char *text, void *data)
 
 /* --------------------------------------------------------------------------------------------- */
 
+static inline char *
+add_to_list_take (char *text, void *data)
+{
+    return listbox_add_item_take (find_list, LISTBOX_APPEND_AT_END, 0, text, data, TRUE);
+}
+
+/* --------------------------------------------------------------------------------------------- */
+
 static inline void
 stop_idle (void *data)
 {
@@ -374,7 +388,7 @@ found_num_update (void)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-get_list_info (char **file, char **dir, gsize * start, gsize * end)
+get_list_info (char **file, char **dir, gsize *start, gsize *end)
 {
     find_match_location_t *location;
 
@@ -459,7 +473,7 @@ find_toggle_enable_content (void)
  */
 
 static cb_ret_t
-find_parm_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, void *data)
+find_parm_callback (Widget *w, Widget *sender, widget_msg_t msg, int parm, void *data)
 {
     /* FIXME: HACK: use first draw of dialog to resolve widget state dependencies.
      * Use this time moment to check input field content. We can't do that in MSG_INIT
@@ -497,7 +511,7 @@ find_parm_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, voi
 
         /* check filename regexp */
         if (!file_pattern_cbox->state && !input_is_empty (in_name)
-            && !find_check_regexp (in_name->buffer))
+            && !find_check_regexp (input_get_ctext (in_name)))
         {
             /* Don't stop the dialog */
             widget_set_state (w, WST_ACTIVE, TRUE);
@@ -507,7 +521,8 @@ find_parm_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, voi
         }
 
         /* check content regexp */
-        if (content_regexp_cbox->state && !content_is_empty && !find_check_regexp (in_with->buffer))
+        if (content_regexp_cbox->state && !content_is_empty
+            && !find_check_regexp (input_get_ctext (in_with)))
         {
             /* Don't stop the dialog */
             widget_set_state (w, WST_ACTIVE, TRUE);
@@ -551,23 +566,23 @@ find_parm_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, voi
  * If the return value is TRUE, then the following holds:
  *
  * start_dir, ignore_dirs, pattern and content contain the information provided by the user.
- * They are newly allocated strings and must be freed when uneeded.
+ * They are newly allocated strings and must be freed when unneeded.
  *
  * start_dir_len is -1 when user entered an absolute path, otherwise it is a length
  * of start_dir (which is absolute). It is used to get a relative pats of find results.
  */
 
 static gboolean
-find_parameters (char **start_dir, ssize_t * start_dir_len,
+find_parameters (WPanel *panel, char **start_dir, ssize_t *start_dir_len,
                  char **ignore_dirs, char **pattern, char **content)
 {
     WGroup *g;
 
     /* Size of the find parameters window */
 #ifdef HAVE_CHARSET
-    const int lines = 18;
+    const int lines = 19;
 #else
-    const int lines = 17;
+    const int lines = 18;
 #endif
     int cols = 68;
 
@@ -576,6 +591,7 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
     /* file name */
     const char *file_name_label = N_("File name:");
     const char *file_recurs_label = N_("&Find recursively");
+    const char *file_follow_symlinks = N_("Follow s&ymlinks");
     const char *file_pattern_label = N_("&Using shell patterns");
 #ifdef HAVE_CHARSET
     const char *file_all_charsets_label = N_("&All charsets");
@@ -608,6 +624,7 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
 
         file_name_label = _(file_name_label);
         file_recurs_label = _(file_recurs_label);
+        file_follow_symlinks = _(file_follow_symlinks);
         file_pattern_label = _(file_pattern_label);
 #ifdef HAVE_CHARSET
         file_all_charsets_label = _(file_all_charsets_label);
@@ -631,11 +648,12 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
     }
 #endif /* ENABLE_NLS */
 
-    /* caclulate dialog width */
+    /* calculate dialog width */
 
     /* widget widths */
     cw = str_term_width1 (file_name_label);
     cw = max (cw, str_term_width1 (file_recurs_label) + 4);
+    cw = max (cw, str_term_width1 (file_follow_symlinks) + 4);
     cw = max (cw, str_term_width1 (file_pattern_label) + 4);
 #ifdef HAVE_CHARSET
     cw = max (cw, str_term_width1 (file_all_charsets_label) + 4);
@@ -718,6 +736,9 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
     recursively_cbox = check_new (y1++, x1, options.find_recurs, file_recurs_label);
     group_add_widget (g, recursively_cbox);
 
+    follow_sym_cbox = check_new (y1++, x1, options.follow_symlinks, file_follow_symlinks);
+    group_add_widget (g, follow_sym_cbox);
+
     file_pattern_cbox = check_new (y1++, x1, options.file_pattern, file_pattern_label);
     group_add_widget (g, file_pattern_cbox);
 
@@ -772,21 +793,21 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
 
     case B_TREE:
         {
-            char *temp_dir;
+            const char *start_cstr;
+            const char *temp_dir;
 
-            temp_dir = in_start->buffer;
-            if (*temp_dir == '\0' || DIR_IS_DOT (temp_dir))
-                temp_dir = g_strdup (vfs_path_as_str (current_panel->cwd_vpath));
+            start_cstr = input_get_ctext (in_start);
+
+            if (input_is_empty (in_start) || DIR_IS_DOT (start_cstr))
+                temp_dir = vfs_path_as_str (panel->cwd_vpath);
             else
-                temp_dir = g_strdup (temp_dir);
+                temp_dir = start_cstr;
 
             if (in_start_dir != INPUT_LAST_TEXT)
                 g_free (in_start_dir);
             in_start_dir = tree_box (temp_dir);
             if (in_start_dir == NULL)
-                in_start_dir = temp_dir;
-            else
-                g_free (temp_dir);
+                in_start_dir = g_strdup (temp_dir);
 
             input_assign_text (in_start, in_start_dir);
 
@@ -807,19 +828,20 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
             options.content_first_hit = content_first_hit_cbox->state;
             options.content_whole_words = content_whole_words_cbox->state;
             options.find_recurs = recursively_cbox->state;
+            options.follow_symlinks = follow_sym_cbox->state;
             options.file_pattern = file_pattern_cbox->state;
             options.file_case_sens = file_case_sens_cbox->state;
             options.skip_hidden = skip_hidden_cbox->state;
             options.ignore_dirs_enable = ignore_dirs_cbox->state;
             g_free (options.ignore_dirs);
-            options.ignore_dirs = g_strdup (in_ignore->buffer);
+            options.ignore_dirs = input_get_text (in_ignore);
 
-            *content = !input_is_empty (in_with) ? g_strdup (in_with->buffer) : NULL;
-            if (!input_is_empty (in_name))
-                *pattern = g_strdup (in_name->buffer);
-            else
+            *content = !input_is_empty (in_with) ? input_get_text (in_with) : NULL;
+            if (input_is_empty (in_name))
                 *pattern = g_strdup (options.file_pattern ? "*" : ".*");
-            *start_dir = !input_is_empty (in_start) ? in_start->buffer : (char *) ".";
+            else
+                *pattern = input_get_text (in_name);
+            *start_dir = (char *) (!input_is_empty (in_start) ? input_get_ctext (in_start) : ".");
             if (in_start_dir != INPUT_LAST_TEXT)
                 g_free (in_start_dir);
             in_start_dir = g_strdup (*start_dir);
@@ -829,8 +851,8 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
 
             if (DIR_IS_DOT (s))
             {
-                *start_dir = g_strdup (vfs_path_as_str (current_panel->cwd_vpath));
-                /* FIXME: is current_panel->cwd_vpath canonicalized? */
+                *start_dir = g_strdup (vfs_path_as_str (panel->cwd_vpath));
+                /* FIXME: is panel->cwd_vpath canonicalized? */
                 /* relative paths will be used in panelization */
                 *start_dir_len = (ssize_t) strlen (*start_dir);
                 g_free (s);
@@ -844,17 +866,16 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
             {
                 /* relative paths will be used in panelization */
                 *start_dir =
-                    mc_build_filename (vfs_path_as_str (current_panel->cwd_vpath), s,
-                                       (char *) NULL);
-                *start_dir_len = (ssize_t) strlen (vfs_path_as_str (current_panel->cwd_vpath));
+                    mc_build_filename (vfs_path_as_str (panel->cwd_vpath), s, (char *) NULL);
+                *start_dir_len = (ssize_t) vfs_path_len (panel->cwd_vpath);
                 g_free (s);
             }
 
             if (!options.ignore_dirs_enable || input_is_empty (in_ignore)
-                || DIR_IS_DOT (in_ignore->buffer))
+                || DIR_IS_DOT (input_get_ctext (in_ignore)))
                 *ignore_dirs = NULL;
             else
-                *ignore_dirs = g_strdup (in_ignore->buffer);
+                *ignore_dirs = input_get_text (in_ignore);
 
             find_save_options ();
 
@@ -862,7 +883,7 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
         }
     }
 
-    dlg_destroy (find_dlg);
+    widget_destroy (WIDGET (find_dlg));
 
     return return_value;
 }
@@ -870,7 +891,7 @@ find_parameters (char **start_dir, ssize_t * start_dir_len,
 /* --------------------------------------------------------------------------------------------- */
 
 static inline void
-push_directory (vfs_path_t * dir)
+push_directory (vfs_path_t *dir)
 {
     g_queue_push_head (&dir_queue, (void *) dir);
 }
@@ -883,13 +904,21 @@ pop_directory (void)
     return (vfs_path_t *) g_queue_pop_head (&dir_queue);
 }
 
+/* --------------------------------------------------------------------------------------------- */
+
+static void
+queue_dir_free (gpointer data)
+{
+    vfs_path_free ((vfs_path_t *) data, TRUE);
+}
+
 /* --------------------------------------------------------------------------------------------- */
 /** Remove all the items from the stack */
 
 static void
 clear_stack (void)
 {
-    g_queue_clear_full (&dir_queue, (GDestroyNotify) vfs_path_free);
+    g_queue_clear_full (&dir_queue, queue_dir_free);
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -897,7 +926,7 @@ clear_stack (void)
 static void
 insert_file (const char *dir, const char *file, gsize start, gsize end)
 {
-    char *tmp_name = NULL;
+    char *tmp_name;
     static char *dirname = NULL;
     find_match_location_t *location;
 
@@ -924,8 +953,7 @@ insert_file (const char *dir, const char *file, gsize start, gsize end)
     location->dir = dirname;
     location->start = start;
     location->end = end;
-    add_to_list (tmp_name, location);
-    g_free (tmp_name);
+    add_to_list_take (tmp_name, location);
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -947,7 +975,7 @@ find_add_match (const char *dir, const char *file, gsize start, gsize end)
 /* --------------------------------------------------------------------------------------------- */
 
 static FindProgressStatus
-check_find_events (WDialog * h)
+check_find_events (WDialog *h)
 {
     Gpm_Event event;
     int c;
@@ -985,51 +1013,33 @@ check_find_events (WDialog * h)
  */
 
 static gboolean
-search_content (WDialog * h, const char *directory, const char *filename)
+search_content (WDialog *h, const char *directory, const char *filename)
 {
     struct stat s;
-    char buffer[BUF_4K];        /* raw input buffer */
-    int file_fd;
+    char buffer[BUF_4K] = "";   /* raw input buffer */
+    int file_fd = -1;
     gboolean ret_val = FALSE;
     vfs_path_t *vpath;
-    struct timeval tv;
-    time_t seconds;
-    suseconds_t useconds;
+    gint64 tv;
     gboolean status_updated = FALSE;
 
     vpath = vfs_path_build_filename (directory, filename, (char *) NULL);
 
-    if (mc_stat (vpath, &s) != 0 || !S_ISREG (s.st_mode))
-    {
-        vfs_path_free (vpath);
-        return FALSE;
-    }
+    if (mc_stat (vpath, &s) == 0 && S_ISREG (s.st_mode))
+        file_fd = mc_open (vpath, O_RDONLY);
 
-    file_fd = mc_open (vpath, O_RDONLY);
-    vfs_path_free (vpath);
+    vfs_path_free (vpath, TRUE);
 
     if (file_fd == -1)
         return FALSE;
 
     /* get time elapsed from last refresh */
-    if (gettimeofday (&tv, NULL) == -1)
-    {
-        tv.tv_sec = 0;
-        tv.tv_usec = 0;
-        last_refresh = tv;
-    }
-    seconds = tv.tv_sec - last_refresh.tv_sec;
-    useconds = tv.tv_usec - last_refresh.tv_usec;
-    if (useconds < 0)
-    {
-        seconds--;
-        useconds += G_USEC_PER_SEC;
-    }
+    tv = g_get_monotonic_time ();
 
-    if (s.st_size >= MIN_REFRESH_FILE_SIZE || seconds > 0 || useconds > MAX_REFRESH_INTERVAL)
+    if (s.st_size >= MIN_REFRESH_FILE_SIZE || (tv - last_refresh) > MAX_REFRESH_INTERVAL)
     {
         g_snprintf (buffer, sizeof (buffer), _("Grepping in %s"), filename);
-        status_update (str_trunc (buffer, WIDGET (h)->cols - 8));
+        status_update (str_trunc (buffer, WIDGET (h)->rect.cols - 8));
         mc_refresh ();
         last_refresh = tv;
         status_updated = TRUE;
@@ -1044,9 +1054,6 @@ search_content (WDialog * h, const char *directory, const char *filename)
         int n_read = 0;
         off_t off = 0;          /* file_fd's offset corresponding to strbuf[0] */
         gboolean found = FALSE;
-        gsize found_len;
-        gsize found_start;
-        char result[BUF_MEDIUM];
         char *strbuf = NULL;    /* buffer for fetched string */
         int strbuf_size = 0;
         int i = -1;             /* compensate for a newline we'll add when we first enter the loop */
@@ -1064,6 +1071,7 @@ search_content (WDialog * h, const char *directory, const char *filename)
         while (!ret_val)
         {
             char ch = '\0';
+            gsize found_len;
 
             off += i + 1;       /* the previous line, plus a newline character */
             i = 0;
@@ -1118,12 +1126,15 @@ search_content (WDialog * h, const char *directory, const char *filename)
             if (!found          /* Search in binary line once */
                 && mc_search_run (search_content_handle, (const void *) strbuf, 0, i, &found_len))
             {
+                gsize found_start;
+                char result[BUF_MEDIUM];
+
                 if (!status_updated)
                 {
                     /* if we add results for a file, we have to ensure that
                        name of this file is shown in status bar */
                     g_snprintf (result, sizeof (result), _("Grepping in %s"), filename);
-                    status_update (str_trunc (result, WIDGET (h)->cols - 8));
+                    status_update (str_trunc (result, WIDGET (h)->rect.cols - 8));
                     mc_refresh ();
                     last_refresh = tv;
                     status_updated = TRUE;
@@ -1185,11 +1196,11 @@ search_content (WDialog * h, const char *directory, const char *filename)
   If dir is relative, this means we're going to add dir to the directory stack.
 **/
 static gboolean
-find_ignore_dir_search (const char *dir)
+find_ignore_dir_search (const char *dir, size_t len)
 {
     if (find_ignore_dirs != NULL)
     {
-        const size_t dlen = strlen (dir);
+        const size_t dlen = len == (size_t) (-1) ? strlen (dir) : len;
         const unsigned char dabs = g_path_is_absolute (dir) ? 1 : 0;
 
         char **ignore_dir;
@@ -1207,7 +1218,7 @@ find_ignore_dir_search (const char *dir)
             switch (iabs | dabs)
             {
             case 0:            /* both paths are relative */
-            case 3:            /* both paths are abolute */
+            case 3:            /* both paths are absolute */
                 /* if ignore dir is not a path  of dir -- skip it */
                 if (strncmp (dir, *ignore_dir, ilen) == 0)
                 {
@@ -1242,7 +1253,7 @@ find_ignore_dir_search (const char *dir)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-find_rotate_dash (const WDialog * h, gboolean show)
+find_rotate_dash (const WDialog *h, gboolean show)
 {
     static size_t pos = 0;
     static const char rotating_dash[4] = "|/-\\";
@@ -1251,7 +1262,7 @@ find_rotate_dash (const WDialog * h, gboolean show)
 
     colors = widget_get_colors (w);
     tty_setcolor (colors[DLG_COLOR_NORMAL]);
-    widget_gotoyx (h, w->lines - 7, w->cols - 4);
+    widget_gotoyx (h, w->rect.lines - 7, w->rect.cols - 4);
     tty_print_char (show ? rotating_dash[pos] : ' ');
     pos = (pos + 1) % sizeof (rotating_dash);
     mc_refresh ();
@@ -1260,11 +1271,12 @@ find_rotate_dash (const WDialog * h, gboolean show)
 /* --------------------------------------------------------------------------------------------- */
 
 static int
-do_search (WDialog * h)
+do_search (WDialog *h)
 {
-    static struct dirent *dp = NULL;
+    static struct vfs_dirent *dp = NULL;
     static DIR *dirp = NULL;
     static char *directory = NULL;
+    static gboolean pop_start_dir = TRUE;
     struct stat tmp_stat;
     gsize bytes_found;
     unsigned short count;
@@ -1278,6 +1290,7 @@ do_search (WDialog * h)
         }
         MC_PTR_FREE (directory);
         dp = NULL;
+        pop_start_dir = TRUE;
         return 1;
     }
 
@@ -1321,32 +1334,37 @@ do_search (WDialog * h)
                         return 0;
                     }
 
-                    /* handle absolute ignore dirs here */
+                    /* The start directory is the first one in the stack (see do_find() below).
+                       Do not apply ignore_dir to it. */
+                    if (pop_start_dir)
                     {
-                        gboolean ok;
-
-                        ok = find_ignore_dir_search (vfs_path_as_str (tmp_vpath));
-                        if (!ok)
-                            break;
+                        pop_start_dir = FALSE;
+                        break;
                     }
 
-                    vfs_path_free (tmp_vpath);
+                    pop_start_dir = FALSE;
+
+                    /* handle absolute ignore dirs here */
+                    if (!find_ignore_dir_search (vfs_path_as_str (tmp_vpath), -1))
+                        break;
+
+                    vfs_path_free (tmp_vpath, TRUE);
                     ignore_count++;
                 }
 
                 g_free (directory);
-                directory = g_strdup (vfs_path_as_str (tmp_vpath));
 
                 if (verbose)
                 {
                     char buffer[BUF_MEDIUM];
 
+                    directory = (char *) vfs_path_as_str (tmp_vpath);
                     g_snprintf (buffer, sizeof (buffer), _("Searching %s"), directory);
-                    status_update (str_trunc (directory, WIDGET (h)->cols - 8));
+                    status_update (str_trunc (directory, WIDGET (h)->rect.cols - 8));
                 }
 
                 dirp = mc_opendir (tmp_vpath);
-                vfs_path_free (tmp_vpath);
+                directory = vfs_path_free (tmp_vpath, FALSE);
             }                   /* while (!dirp) */
 
             /* skip invalid filenames */
@@ -1370,23 +1388,28 @@ do_search (WDialog * h)
             if (options.find_recurs && (directory != NULL))
             {                   /* Can directory be NULL ? */
                 /* handle relative ignore dirs here */
-                if (options.ignore_dirs_enable && find_ignore_dir_search (dp->d_name))
+                if (options.ignore_dirs_enable && find_ignore_dir_search (dp->d_name, dp->d_len))
                     ignore_count++;
                 else
                 {
                     vfs_path_t *tmp_vpath;
+                    int stat_res;
 
                     tmp_vpath = vfs_path_build_filename (directory, dp->d_name, (char *) NULL);
 
-                    if (mc_lstat (tmp_vpath, &tmp_stat) == 0 && S_ISDIR (tmp_stat.st_mode))
+                    if (options.follow_symlinks)
+                        stat_res = mc_stat (tmp_vpath, &tmp_stat);
+                    else
+                        stat_res = mc_lstat (tmp_vpath, &tmp_stat);
+
+                    if (stat_res == 0 && S_ISDIR (tmp_stat.st_mode))
                         push_directory (tmp_vpath);
                     else
-                        vfs_path_free (tmp_vpath);
+                        vfs_path_free (tmp_vpath, TRUE);
                 }
             }
 
-            search_ok = mc_search_run (search_file_handle, dp->d_name,
-                                       0, strlen (dp->d_name), &bytes_found);
+            search_ok = mc_search_run (search_file_handle, dp->d_name, 0, dp->d_len, &bytes_found);
 
             if (search_ok)
             {
@@ -1451,7 +1474,7 @@ find_do_view_edit (gboolean unparsed_view, gboolean edit, char *dir, char *file,
     else
         view_file_at_line (fullname_vpath, unparsed_view, use_internal_view, line, search_start,
                            search_end);
-    vfs_path_free (fullname_vpath);
+    vfs_path_free (fullname_vpath, TRUE);
 }
 
 /* --------------------------------------------------------------------------------------------- */
@@ -1474,9 +1497,9 @@ view_edit_currently_selected_file (gboolean unparsed_view, gboolean edit)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-find_calc_button_locations (const WDialog * h, gboolean all_buttons)
+find_calc_button_locations (const WDialog *h, gboolean all_buttons)
 {
-    const int cols = CONST_WIDGET (h)->cols;
+    const int cols = CONST_WIDGET (h)->rect.cols;
 
     int l1, l2;
 
@@ -1500,7 +1523,7 @@ find_calc_button_locations (const WDialog * h, gboolean all_buttons)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-find_adjust_header (WDialog * h)
+find_adjust_header (WDialog *h)
 {
     char title[BUF_MEDIUM];
     int title_len;
@@ -1512,10 +1535,10 @@ find_adjust_header (WDialog * h)
         g_snprintf (title, sizeof (title), _("Find File: \"%s\""), find_pattern);
 
     title_len = str_term_width1 (title);
-    if (title_len > WIDGET (h)->cols - 6)
+    if (title_len > WIDGET (h)->rect.cols - 6)
     {
         /* title is too wide, truncate it */
-        title_len = WIDGET (h)->cols - 6;
+        title_len = WIDGET (h)->rect.cols - 6;
         title_len = str_column_to_pos (title, title_len);
         title_len -= 3;         /* reserve space for three dots */
         title_len = str_offset_to_pos (title, title_len);
@@ -1529,25 +1552,26 @@ find_adjust_header (WDialog * h)
 /* --------------------------------------------------------------------------------------------- */
 
 static void
-find_relocate_buttons (const WDialog * h, gboolean all_buttons)
+find_relocate_buttons (const WDialog *h, gboolean all_buttons)
 {
     size_t i;
 
     find_calc_button_locations (h, all_buttons);
 
     for (i = 0; i < fbuts_num; i++)
-        fbuts[i].button->x = CONST_WIDGET (h)->x + fbuts[i].x;
+        fbuts[i].button->rect.x = CONST_WIDGET (h)->rect.x + fbuts[i].x;
 }
 
 /* --------------------------------------------------------------------------------------------- */
 
 static cb_ret_t
-find_resize (WDialog * h)
+find_resize (WDialog *h)
 {
     Widget *w = WIDGET (h);
-    WRect r;
+    WRect r = w->rect;
 
-    rect_init (&r, w->y, w->x, LINES - 4, COLS - 16);
+    r.lines = LINES - 4;
+    r.cols = COLS - 16;
     dlg_default_callback (w, NULL, MSG_RESIZE, 0, &r);
     find_adjust_header (h);
     find_relocate_buttons (h, TRUE);
@@ -1558,7 +1582,7 @@ find_resize (WDialog * h)
 /* --------------------------------------------------------------------------------------------- */
 
 static cb_ret_t
-find_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, void *data)
+find_callback (Widget *w, Widget *sender, widget_msg_t msg, int parm, void *data)
 {
     WDialog *h = DIALOG (w);
 
@@ -1596,7 +1620,7 @@ find_callback (Widget * w, Widget * sender, widget_msg_t msg, int parm, void *da
 /** Handles the Stop/Start button in the find window */
 
 static int
-start_stop (WButton * button, int action)
+start_stop (WButton *button, int action)
 {
     Widget *w = WIDGET (button);
 
@@ -1619,7 +1643,7 @@ start_stop (WButton * button, int action)
 /** Handle view command, when invoked as a button */
 
 static int
-find_do_view_file (WButton * button, int action)
+find_do_view_file (WButton *button, int action)
 {
     (void) button;
     (void) action;
@@ -1632,7 +1656,7 @@ find_do_view_file (WButton * button, int action)
 /** Handle edit command, when invoked as a button */
 
 static int
-find_do_edit_file (WButton * button, int action)
+find_do_edit_file (WButton *button, int action)
 {
     (void) button;
     (void) action;
@@ -1681,11 +1705,11 @@ setup_gui (void)
     y = 2;
     find_list = listbox_new (y, 2, lines - 10, cols - 4, FALSE, NULL);
     group_add_widget_autopos (g, find_list, WPOS_KEEP_ALL, NULL);
-    y += WIDGET (find_list)->lines;
+    y += WIDGET (find_list)->rect.lines;
 
     group_add_widget_autopos (g, hline_new (y++, -1, -1), WPOS_KEEP_BOTTOM, NULL);
 
-    found_num_label = label_new (y++, 4, "");
+    found_num_label = label_new (y++, 4, NULL);
     group_add_widget_autopos (g, found_num_label, WPOS_KEEP_BOTTOM, NULL);
 
     status_label = label_new (y++, 4, _("Searching"));
@@ -1757,17 +1781,19 @@ run_process (void)
 static void
 kill_gui (void)
 {
-    widget_idle (WIDGET (find_dlg), FALSE);
-    dlg_destroy (find_dlg);
+    Widget *w = WIDGET (find_dlg);
+
+    widget_idle (w, FALSE);
+    widget_destroy (w);
 }
 
 /* --------------------------------------------------------------------------------------------- */
 
 static int
-do_find (const char *start_dir, ssize_t start_dir_len, const char *ignore_dirs,
+do_find (WPanel *panel, const char *start_dir, ssize_t start_dir_len, const char *ignore_dirs,
          char **dirname, char **filename)
 {
-    int return_value = 0;
+    int return_value;
     char *dir_tmp = NULL, *file_tmp = NULL;
 
     setup_gui ();
@@ -1783,26 +1809,26 @@ do_find (const char *start_dir, ssize_t start_dir_len, const char *ignore_dirs,
 
     get_list_info (&file_tmp, &dir_tmp, NULL, NULL);
 
-    if (dir_tmp)
+    if (dir_tmp != NULL)
         *dirname = g_strdup (dir_tmp);
-    if (file_tmp)
+    if (file_tmp != NULL)
         *filename = g_strdup (file_tmp);
 
-    if (return_value == B_PANELIZE && *filename)
+    if (return_value == B_PANELIZE && *filename != NULL)
     {
-        int i;
         struct stat st;
         GList *entry;
-        dir_list *list = &current_panel->dir;
+        dir_list *list = &panel->dir;
         char *name = NULL;
+        gboolean ok = TRUE;
 
-        panel_clean_dir (current_panel);
+        panel_clean_dir (panel);
         dir_list_init (list);
 
-        for (i = 0, entry = listbox_get_first_link (find_list); entry != NULL;
-             i++, entry = g_list_next (entry))
+        for (entry = listbox_get_first_link (find_list); entry != NULL && ok;
+             entry = g_list_next (entry))
         {
-            const char *lc_filename = NULL;
+            const char *lc_filename;
             WLEntry *le = LENTRY (entry->data);
             find_match_location_t *location = le->data;
             char *p;
@@ -1818,53 +1844,37 @@ do_find (const char *start_dir, ssize_t start_dir_len, const char *ignore_dirs,
 
             name = mc_build_filename (location->dir, lc_filename, (char *) NULL);
             /* skip initial start dir */
-            if (start_dir_len < 0)
-                p = name;
-            else
-            {
-                p = name + (size_t) start_dir_len;
-                if (IS_PATH_SEP (*p))
-                    p++;
-            }
+            p = name;
+            if (start_dir_len > 0)
+                p += (size_t) start_dir_len;
+            if (IS_PATH_SEP (*p))
+                p++;
 
             if (!handle_path (p, &st, &link_to_dir, &stale_link))
             {
                 g_free (name);
                 continue;
             }
-            /* Need to grow the *list? */
-            if (list->len == list->size && !dir_list_grow (list, DIR_LIST_RESIZE_STEP))
-            {
-                g_free (name);
-                break;
-            }
 
             /* don't add files more than once to the panel */
             if (!content_is_empty && list->len != 0
-                && strcmp (list->list[list->len - 1].fname, p) == 0)
+                && strcmp (list->list[list->len - 1].fname->str, p) == 0)
             {
                 g_free (name);
                 continue;
             }
 
-            list->list[list->len].fnamelen = strlen (p);
-            list->list[list->len].fname = g_strndup (p, list->list[list->len].fnamelen);
-            list->list[list->len].f.marked = 0;
-            list->list[list->len].f.link_to_dir = link_to_dir ? 1 : 0;
-            list->list[list->len].f.stale_link = stale_link ? 1 : 0;
-            list->list[list->len].f.dir_size_computed = 0;
-            list->list[list->len].st = st;
-            list->list[list->len].sort_key = NULL;
-            list->list[list->len].second_sort_key = NULL;
-            list->len++;
+            ok = dir_list_append (list, p, &st, link_to_dir, stale_link);
+
             g_free (name);
+
             if ((list->len & 15) == 0)
                 rotate_dash (TRUE);
         }
 
-        current_panel->is_panelized = TRUE;
-        panelize_absolutize_if_needed (current_panel);
-        panelize_save_panel (current_panel);
+        panel->is_panelized = TRUE;
+        panel_panelize_absolutize_if_needed (panel);
+        panel_panelize_save (panel);
     }
 
     kill_gui ();
@@ -1880,7 +1890,7 @@ do_find (const char *start_dir, ssize_t start_dir_len, const char *ignore_dirs,
 /* --------------------------------------------------------------------------------------------- */
 
 void
-find_cmd (void)
+find_cmd (WPanel *panel)
 {
     char *start_dir = NULL, *ignore_dirs = NULL;
     ssize_t start_dir_len;
@@ -1888,7 +1898,7 @@ find_cmd (void)
     find_pattern = NULL;
     content_pattern = NULL;
 
-    while (find_parameters (&start_dir, &start_dir_len,
+    while (find_parameters (panel, &start_dir, &start_dir_len,
                             &ignore_dirs, &find_pattern, &content_pattern))
     {
         char *filename = NULL, *dirname = NULL;
@@ -1898,15 +1908,14 @@ find_cmd (void)
 
         if (find_pattern[0] != '\0')
         {
-            last_refresh.tv_sec = 0;
-            last_refresh.tv_usec = 0;
+            last_refresh = 0;
 
             is_start = FALSE;
 
             if (!content_is_empty && !str_is_valid_string (content_pattern))
                 MC_PTR_FREE (content_pattern);
 
-            v = do_find (start_dir, start_dir_len, ignore_dirs, &dirname, &filename);
+            v = do_find (panel, start_dir, start_dir_len, ignore_dirs, &dirname, &filename);
         }
 
         g_free (start_dir);
@@ -1920,20 +1929,28 @@ find_cmd (void)
                 vfs_path_t *dirname_vpath;
 
                 dirname_vpath = vfs_path_from_str (dirname);
-                do_cd (dirname_vpath, cd_exact);
-                vfs_path_free (dirname_vpath);
+                panel_cd (panel, dirname_vpath, cd_exact);
+                vfs_path_free (dirname_vpath, TRUE);
+
                 if (filename != NULL)
-                    try_to_select (current_panel,
-                                   filename + (content_pattern != NULL
-                                               ? strchr (filename + 4, ':') - filename + 1 : 4));
+                {
+                    size_t offset;
+
+                    if (content_pattern == NULL)
+                        offset = 4;
+                    else
+                        offset = strchr (filename + 4, ':') - filename + 1;
+
+                    panel_set_current_by_name (panel, filename + offset);
+                }
             }
             else if (filename != NULL)
             {
                 vfs_path_t *filename_vpath;
 
                 filename_vpath = vfs_path_from_str (filename);
-                do_cd (filename_vpath, cd_exact);
-                vfs_path_free (filename_vpath);
+                panel_cd (panel, filename_vpath, cd_exact);
+                vfs_path_free (filename_vpath, TRUE);
             }
         }
 
@@ -1946,8 +1963,8 @@ find_cmd (void)
 
         if (v == B_PANELIZE)
         {
-            panel_re_sort (current_panel);
-            try_to_select (current_panel, NULL);
+            panel_re_sort (panel);
+            panel_set_current_by_name (panel, NULL);
             break;
         }
     }
